"""Core Swimlane client class"""

import logging

import requests
from pyuri import URI
from requests.compat import json
from requests.packages import urllib3
from requests.structures import CaseInsensitiveDict
from six.moves.urllib.parse import urljoin

from swimlane.core.adapters import GroupAdapter, UserAdapter, AppAdapter, HelperAdapter
from swimlane.core.cache import ResourcesCache
from swimlane.core.resolver import SwimlaneResolver
from swimlane.core.resources.usergroup import User
from swimlane.exceptions import SwimlaneHTTP400Error, InvalidSwimlaneProductVersion
from swimlane.utils.version import get_package_version, compare_versions

# Disable insecure request warnings
urllib3.disable_warnings()

logger = logging.getLogger(__name__)

# pylint: disable=invalid-name
_lib_full_version = get_package_version()
_lib_major_version, _lib_minor_version = _lib_full_version.split('.')[0:2]


class Swimlane(object):
    """Swimlane API client

    Core class used throughout library for all API requests and server interactions

    Args:
        host (str): Full RFC-1738 URL pointing to Swimlane host. Defaults will be provided for all parts
        username (str): Authentication username
        password (str): Authentication password
        verify_ssl (bool): Verify SSL (ignored on HTTP). Disable to use self-signed certificates
        default_timeout (int): Default request connect and read timeout in seconds for all requests
        verify_server_version (bool): Verify server version has same major version as client package. May require
            additional requests, set False to disable check
        resource_cache_size (int): Maximum number of each resource type to keep in memory cache. Set 0 to disable
            caching. Disabled by default

    Attributes:
        host (pyuri.URI): Full RFC-1738 URL pointing to Swimlane host
        apps (AppAdapter): :class:`~swimlane.core.adapters.app.AppAdapter` configured for current Swimlane instance
        users (UserAdapter): :class:`~swimlane.core.adapters.usergroup.UserAdapter` configured for current
            Swimlane instance
        groups (GroupAdapter): :class:`~swimlane.core.adapters.usergroup.GroupAdapter` configured for current
            Swimlane instance
        resources_cache (ResourcesCache): Cache checked by all supported adapters for current Swimlane instance

    Examples:

        ::

            # Establish connection
            swimlane = Swimlane(
                'https://192.168.1.1',
                'username',
                'password',
                verify_ssl=False
            )

            # Retrieve an app
            app = swimlane.apps.get(name='Target App')

    """

    _api_root = '/api/'

    def __init__(
            self,
            host,
            username,
            password,
            verify_ssl=True,
            default_timeout=60,
            verify_server_version=True,
            resource_cache_size=0
    ):
        self.host = URI(host)
        self.host.scheme = self.host.scheme.lower() or 'https'
        self.host.path = None

        self.resources_cache = ResourcesCache(resource_cache_size)

        self.__settings = None
        self.__user = None

        self._default_timeout = default_timeout

        self._session = requests.Session()
        self._session.verify = verify_ssl
        self._session.auth = SwimlaneAuth(
            self,
            username,
            password,
            verify_ssl
        )

        self.apps = AppAdapter(self)
        self.users = UserAdapter(self)
        self.groups = GroupAdapter(self)
        self.helpers = HelperAdapter(self)

        if verify_server_version:
            self.__verify_server_version()

    def __verify_server_version(self):
        """Verify connected to supported server product version

        Notes:
            Logs warning if connecting to a newer minor server version

        Raises:
            swimlane.exceptions.InvalidServerVersion: If server major version is higher than package major version
        """
        if compare_versions('.'.join([_lib_major_version, _lib_minor_version]), self.product_version) > 0:
            logger.warning('Client version {} connecting to server with newer minor release {}.'.format(
                _lib_full_version,
                self.product_version
            ))

        if compare_versions(_lib_major_version, self.product_version) != 0:
            raise InvalidSwimlaneProductVersion(
                self,
                '{}.0'.format(_lib_major_version),
                '{}.0'.format(str(int(_lib_major_version) + 1))
            )

    def __repr__(self):
        return '<{cls}: {user} @ {host} v{version}>'.format(
            cls=self.__class__.__name__,
            user=self.user,
            host=self.host,
            version=self.version
        )

    def request(self, method, api_endpoint, **kwargs):
        """Wrapper for underlying :class:`requests.Session`

        Handles generating full API URL, session reuse and auth, request defaults, and invalid response status codes

        Used throughout library as the core underlying request/response method for all interactions with server

        Args:
            method (str): Request method (get, post, put, etc.)
            api_endpoint (str): Portion of URL matching API endpoint route as listed in platform /docs help page
            **kwargs (dict): Remaining arguments passed through to actual request call

        Notes:
            All other provided kwargs are passed to underlying :meth:`requests.Session.request()` call

        Raises:
            swimlane.exceptions.SwimlaneHTTP400Error: On 400 responses with additional context about the exception
            requests.HTTPError: Any other 4xx/5xx HTTP responses

        Returns:
            requests.Response: Successful response instances

        Examples:

            Request and parse server settings endpoint response

            >>> server_settings = swimlane.request('get', 'settings').json()
        """
        while api_endpoint.startswith('/'):
            api_endpoint = api_endpoint[1:]

        # Ensure a timeout is set
        kwargs.setdefault('timeout', self._default_timeout)

        # Manually grab and dump json data to have full control over serialization
        # Emulate default requests behavior
        json_data = kwargs.pop('json', None)
        if json_data is not None:
            headers = CaseInsensitiveDict(kwargs.get('headers', {}))
            headers.setdefault('Content-Type', 'application/json')
            kwargs['headers'] = headers

            kwargs['data'] = json.dumps(json_data, sort_keys=True, separators=(',', ':'))

        response = self._session.request(method, urljoin(str(self.host) + self._api_root, api_endpoint), **kwargs)

        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            if error.response.status_code == 400:
                raise SwimlaneHTTP400Error(error)
            else:
                raise error

        return response

    @property
    def settings(self):
        """Retrieve and cache settings from server"""
        if not self.__settings:
            self.__settings = self.request('get', 'settings').json()
        return self.__settings

    @property
    def version(self):
        """Full Swimlane version, <product_version>+<build_version>+<build_number>"""
        return self.settings['apiVersion']

    @property
    def product_version(self):
        """Swimlane product version"""
        version_separator = '+'
        if version_separator in self.version:
            # Post product/build version separation
            return self.version.split(version_separator)[0]
        # Pre product/build version separation
        return self.version.split('-')[0]

    @property
    def build_version(self):
        """Swimlane semantic build version

        Falls back to product version in pre-2.18 releases
        """
        version_separator = '+'
        if version_separator in self.version:
            # Post product/build version separation
            return self.version.split(version_separator)[1]
        # Pre product/build version separation
        return self.product_version

    @property
    def build_number(self):
        """Swimlane build number"""
        version_separator = '+'
        if version_separator in self.version:
            # Post product/build version separation
            return self.version.split(version_separator)[2]
        # Pre product/build version separation
        return self.version.split('-')[1]

    @property
    def user(self):
        """User record instance for authenticated user"""
        return self._session.auth.user


class SwimlaneAuth(SwimlaneResolver):
    """Handles authentication for all requests"""

    def __init__(self, swimlane, username, password, verify_ssl=True):
        super(SwimlaneAuth, self).__init__(swimlane)

        self.user, self._login_headers = self.authenticate(username, password, verify_ssl)

    def __call__(self, request):

        request.headers.update(self._login_headers)

        return request

    def authenticate(self, username, password, verify_ssl=True):
        """Send login request and return User instance and login headers"""
        # Explicitly provide verify_ssl argument, appears to not consistently be acknowledged across versions during
        # initial setup for auth
        resp = self._swimlane.request(
            'post',
            'user/login',
            json={
                'userName': username,
                'password': password,
                'domain': ''
            },
            verify=verify_ssl
        )
        json_content = resp.json()

        # Check for token in response content
        token = json_content.pop('token', None)

        if token is None:
            # Legacy cookie authentication (2.13-)
            headers = {'Cookie': ';'.join(
                ["%s=%s" % cookie for cookie in resp.cookies.items()]
            )}
        else:
            # JWT auth (2.14+)
            headers = {
                'Authorization': 'Bearer {}'.format(token)
            }

        # User

        user = User(self._swimlane, _user_raw_from_login_content(json_content))

        return user, headers


def _user_raw_from_login_content(login_content):
    """Returns a User instance with appropriate raw data parsed from login response content"""
    matching_keys = [
        'displayName',
        'lastLogin',
        'active',
        'name',
        'isMe',
        'lastPasswordChangedDate',
        'passwordResetRequired',
        'groups',
        'roles',
        'email',
        'isAdmin',
        'createdDate',
        'modifiedDate',
        'createdByUser',
        'modifiedByUser',
        'userName',
        'id',
        'disabled'
    ]

    raw_data = {
        '$type': User._type,
    }

    for key in matching_keys:
        if key in login_content:
            raw_data[key] = login_content[key]

    return raw_data



from __future__ import absolute_import

from datetime import date, datetime, time, timedelta

import math
import pendulum

from .base import Field

UTC = pendulum.timezone('UTC')


class DatetimeField(Field):

    field_type = 'Core.Models.Fields.Date.DateField, Core'

    datetime_format = '%Y-%m-%dT%H:%M:%S.%fZ'

    _type_date = 'date'
    _type_time = 'time'
    _type_interval = 'timespan'

    # All others default to datetime
    _input_type_map = {
        _type_interval: [timedelta],
        _type_date: [datetime, date],
        _type_time: [datetime, time]
    }

    def __init__(self, *args, **kwargs):
        super(DatetimeField, self).__init__(*args, **kwargs)

        # Determine supported_types after inspecting input subtype
        self.supported_types = self._input_type_map.get(self.input_type, [datetime])

    def _set(self, value):
        # Force to appropriate Pendulum instance for consistency
        if value is not None:
            if self.input_type == self._type_interval:
                # Pendulum interval
                value = pendulum.interval.instance(value)
            else:
                if self.input_type == self._type_date:
                    # Pendulum date
                    if isinstance(value, date):
                        value = pendulum.combine(value, pendulum.time(0))
                elif self.input_type == self._type_time:
                    # Pendulum time
                    if isinstance(value, time):
                        value = pendulum.combine(pendulum.date.today(), value)

                # Convert to Pendulum instance in UTC
                value = UTC.convert(pendulum.instance(value))
                # Drop nanosecond precision to match Mongo precision
                value = value.microsecond_(int(math.floor(value.microsecond / 1000) * 1000))

        return super(DatetimeField, self)._set(value)

    def cast_to_python(self, value):
        if value is not None:
            if self.input_type == self._type_interval:
                value = pendulum.interval(milliseconds=int(value))
            else:
                value = pendulum.parse(value)

        return value

    def get_python(self):
        """Coerce to best date type representation for the field subtype"""
        value = super(DatetimeField, self).get_python()

        if value is not None:
            # Handle subtypes with matching Pendulum types
            if self.input_type == self._type_time:
                value = value.time()
            if self.input_type == self._type_date:
                value = value.date()

        return value

    @classmethod
    def format_datetime(cls, target_datetime):
        """Format datetime as expected by Swimlane API"""
        return pendulum.timezone('UTC').convert(target_datetime).strftime(cls.datetime_format)

    def cast_to_swimlane(self, value):
        """Return datetimes formatted as expected by API and timespans as millisecond epochs"""
        if value is None:
            return value

        if self.input_type == self._type_interval:
            return value.in_seconds() * 1000

        return self.format_datetime(value)



"""Creates dist/swimlane-python-offline-installer-win_amd64-<py_version>.zip offline driver installer for windows"""
import glob
import os
import tempfile
import shutil
from zipfile import ZipFile

import sys
from pip import main as pipmain

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

ALL_DEPS_DIR = tempfile.mkdtemp()

# Python configs
PY_VERSION = '{}{}'.format(sys.version_info.major, sys.version_info.minor)
PY_PLATFORM = 'win_amd64'

# Build local wheel
print('Building swimlane wheel')
os.chdir(os.path.join(ROOT_DIR, '..'))
pipmain(['wheel', '--no-deps', '-w', ALL_DEPS_DIR, '.'])
swimlane_whl = glob.glob(os.path.join(ALL_DEPS_DIR, 'swimlane-*'))[0]
swimlane_version = swimlane_whl.split('/')[-1].split('-')[1]
print('Built swimlane version ' + swimlane_version)

# Collect and build all dependencies

print('Downloading all dependencies to {}'.format(ALL_DEPS_DIR))

os.chdir(ALL_DEPS_DIR)

# Initial pass to recursively grab all deps, regardless of platform support
pipmain(['download', '-r', os.path.join(ROOT_DIR, '..', 'requirements.txt')])

ZIPAPP_DIR = tempfile.mkdtemp()

REAL_DEPS_DIR = os.path.join(ZIPAPP_DIR, 'deps')
os.mkdir(REAL_DEPS_DIR)
os.chdir(REAL_DEPS_DIR)

print('Collecting real dependencies to {}'.format(REAL_DEPS_DIR))

for f in os.listdir(ALL_DEPS_DIR):

    filename = os.path.join(ALL_DEPS_DIR, f)

    if f.endswith('.tar.gz'):
        f = f[:-7]

    name, version = f.rsplit('-', 4)[:2]
    package = '=='.join([name, version])

    # Try to download Windows-platform wheel
    if pipmain([
        'download',
        '--no-deps',
        '--platform={}'.format(PY_PLATFORM),
        '--python-version={}'.format(PY_VERSION),
        '--abi=cp{}m'.format(PY_VERSION),
        '--only-binary=:all:',
        package
    ]):
        # Non-zero exit code, package not available as a Windows wheel, just copy in tar file as dep
        shutil.copy(filename, REAL_DEPS_DIR)

print('Collected dependencies to {}'.format(REAL_DEPS_DIR))

# Create self-extracting zip installer

print('Creating self-installing zip')
DIST_DIR = os.path.join(ROOT_DIR, '..', 'dist')
if not os.path.isdir(DIST_DIR):
    os.mkdir(DIST_DIR)

zippath = os.path.realpath(os.path.join(
    DIST_DIR,
    'swimlane-python-{version}-offline-installer-{platform}-py{py_version}.pyz'.format(
        version=swimlane_version,
        platform=PY_PLATFORM,
        py_version=PY_VERSION
    )
))

if os.path.exists(zippath):
    os.remove(zippath)

with ZipFile(zippath, 'w') as zipf:
    zipf.write(os.path.join(ROOT_DIR, '__main__.py'), '__main__.py')

    for f in os.listdir(REAL_DEPS_DIR):
        print('Adding file "{}"'.format(f))
        zipf.write(os.path.join(REAL_DEPS_DIR, f), os.path.join('deps', f))

print('Offline installer written to {}'.format(zippath))



from setuptools import setup, find_packages


def parse_requirements(requirement_file):
    with open(requirement_file) as f:
        return f.readlines()


with open('./README.rst') as f:
    long_description = f.read()


setup(
    name="swimlane",
    author="Swimlane LLC",
    author_email="info@swimlane.com",
    url="https://github.com/swimlane/swimlane-python",
    packages=find_packages(exclude=('tests', 'tests.*')),
    description="Python driver for the Swimlane API",
    long_description=long_description,
    license='AGPLv3',
    install_requires=parse_requirements('./requirements.txt'),
    setup_requires=[
        'setuptools_scm',
        'pytest-runner'
    ],
    use_scm_version=True,
    tests_require=parse_requirements('./test-requirements.txt'),
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ]
)



import mock
import pytest


def test_get(mock_swimlane, mock_app):
    mock_response = mock.MagicMock()

    with mock.patch.object(mock_swimlane, 'request', return_value=mock_response):

        mock_response.status_code = 200
        mock_response.json.return_value = mock_app._raw

        assert mock_swimlane.apps.get(id=mock_app.id).id == mock_app.id

        mock_response.json.return_value = [mock_app._raw]

        assert mock_swimlane.apps.get(name=mock_app.name).id == mock_app.id
        mock_response.json.return_value = []

        with pytest.raises(ValueError):
            mock_swimlane.apps.get(name=mock_app.name)

        mock_response.status_code = 204

        with pytest.raises(ValueError):
            mock_swimlane.apps.get(id=mock_app.id)


@pytest.mark.parametrize('kwargs', [
    {'unknown_arg': 'arg'},
    {'name': 'name', 'id': 'id'},
    {}
])
def test_invalid_args(mock_swimlane, kwargs):
    with pytest.raises(TypeError):
        mock_swimlane.apps.get(**kwargs)



import six

from swimlane.exceptions import ValidationError
from .base import MultiSelectField


class ValuesListField(MultiSelectField):

    field_type = (
        'Core.Models.Fields.ValuesListField, Core',
        'Core.Models.Fields.ValuesList.ValuesListField, Core'
    )
    supported_types = six.string_types

    def __init__(self, *args, **kwargs):
        """Map names to IDs for use in field rehydration"""
        super(ValuesListField, self).__init__(*args, **kwargs)
        self.selection_to_id_map = {f['name']: f['id'] for f in self.field_definition['values']}

    def validate_value(self, value):
        """Validate provided value is one of the valid options"""
        super(ValuesListField, self).validate_value(value)

        if value is not None:
            if value not in self.selection_to_id_map:
                raise ValidationError(
                    self.record,
                    'Field "{}" invalid value "{}". Valid options: {}'.format(
                        self.name,
                        value,
                        ', '.join(self.selection_to_id_map.keys())
                    )
                )

    def cast_to_python(self, value):
        """Store actual value as internal representation"""
        if value is not None:
            value = value['value']

        return value

    def cast_to_swimlane(self, value):
        """Rehydrate value back as full JSON representation"""
        if value is None:
            return value

        return {
            '$type': 'Core.Models.Record.ValueSelection, Core',
            'id': self.selection_to_id_map[value],
            'value': value
        }

    def cast_to_report(self, value):
        value = super(ValuesListField, self).cast_to_report(value)

        if value:
            return value['id']



import copy

import pytest

from swimlane.exceptions import UnknownField


def test_repr(mock_app):
    assert repr(mock_app) == '<App: RSA Alerts (RA)>'


def test_get_field_definitions(mock_app):
    """Test retrieving field definitions by name or id and UnknownField error with recommendation"""
    field_def = mock_app.get_field_definition_by_name('Numeric')
    assert field_def['name'] == 'Numeric'

    assert field_def == mock_app.get_field_definition_by_id(field_def['id'])

    # Test UnknownField metadata
    # Suggestions for potential typo
    try:
        mock_app.get_field_definition_by_name('Muneric')
    except UnknownField as error:
        assert error.app is mock_app
        assert error.field_name == 'Muneric'
        assert error.similar_field_names == ['Numeric']
        assert 'Numeric' in str(error)
    else:
        raise RuntimeError

    # Same behavior for get_by_id
    try:
        mock_app.get_field_definition_by_id('aqkf3')
    except UnknownField as error:
        assert error.app is mock_app
        assert error.field_name == 'aqkf3'
        assert error.similar_field_names == ['aqkg3', 'ayqk6', 'aqc6k']
    else:
        raise RuntimeError


def test_resolve_field_name(mock_app):
    """Test that fields keys + names resolve to field name or None if not found"""
    # Field name
    assert mock_app.resolve_field_name('Action') == 'Action'
    # Key
    assert mock_app.resolve_field_name('action-key') == 'Action'
    # Missing
    assert mock_app.resolve_field_name('unknown field name/key') is None


def test_equality(mock_app):
    """Test equality and inequality of Apps by name"""
    app_clone = copy.copy(mock_app)
    assert app_clone == mock_app
    assert not app_clone != mock_app

    app_clone.name = mock_app.name + 'test'
    assert app_clone != mock_app
    assert not app_clone == mock_app


def test_comparison(mock_app, mock_record):
    with pytest.raises(TypeError):
        mock_app < mock_record

    app_clone = copy.copy(mock_app)
    app_clone.name = mock_app.name + 'test'
    assert mock_app < app_clone
    assert app_clone > mock_app
